###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024 - Yan Georget
###############################################################################
from numba import njit  # type: ignore
from numpy.typing import NDArray

from nucs.constants import MAX, MIN


@njit(cache=True)
def first_not_instantiated_var_heuristic(
    decision_domains: NDArray, shr_domains_stack: NDArray, stacks_top: NDArray, params: NDArray
) -> int:
    """
    Chooses the first non-instantiated shared domain.
    :param decision_domains: the indices of a subset of the shared domains
    :param shr_domains_stack: the stack of shared domains
    :param stacks_top: the index of the top of the stacks as a Numpy array
    :param params: a two-dimensional parameters array, unused here
    :return: the index of the shared domain
    """
    top = stacks_top[0]
    for dom_idx in decision_domains:
        if shr_domains_stack[top, dom_idx, MIN] < shr_domains_stack[top, dom_idx, MAX]:
            return dom_idx
    return -1  # cannot happen



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024 - Yan Georget
###############################################################################
import numpy as np
from numba import njit  # type: ignore
from numpy.typing import NDArray

from nucs.constants import EVENT_MASK_MIN_MAX, MAX, MIN, PROP_CONSISTENCY, PROP_INCONSISTENCY


def get_complexity_max_eq(n: int, parameters: NDArray) -> float:
    """
    Returns the time complexity of the propagator as a float.
    :param n: the number of variables
    :param parameters: the parameters, unused here
    :return: a float
    """
    return 3 * n


def get_triggers_max_eq(n: int, parameters: NDArray) -> NDArray:
    """
    Returns the triggers for this propagator.
    :param n: the number of variables
    :param parameters: the parameters, unused here
    :return: an array of triggers
    """
    return np.full(n, dtype=np.uint8, fill_value=EVENT_MASK_MIN_MAX)


@njit(cache=True)
def compute_domains_max_eq(domains: NDArray, parameters: NDArray) -> int:
    """
    Implements Max_i x_i = x_{n-1}.
    :param domains: the domains of the variables, x is an alias for domains
    :param parameters: unused here
    :return: the status of the propagation (consistency, inconsistency or entailement) as an int
    """
    x = domains[:-1]
    y = domains[-1]
    y[MIN] = max(y[MIN], np.max(x[:, MIN]))
    y[MAX] = min(y[MAX], np.max(x[:, MAX]))
    if y[MIN] > y[MAX]:
        return PROP_INCONSISTENCY
    candidates_nb = 0
    candidate_idx = -1
    for i in range(len(x)):
        if x[i, MAX] >= y[MAX]:
            x[i, MAX] = y[MAX]
            candidate_idx = i
            candidates_nb += 1
    if candidates_nb == 1:
        x[candidate_idx, MIN] = y[MIN]
    return PROP_CONSISTENCY



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024 - Yan Georget
###############################################################################
from typing import Dict

from numpy.typing import NDArray

from nucs.problems.problem import Problem
from nucs.propagators.propagators import ALG_AFFINE_EQ, ALG_ALLDIFFERENT

A, B, C, D, E, F, G, H, I, J, K, L, M, N, O, P, Q, R, S, T, U, V, W, X, Y, Z = tuple(range(26))


class AlphaProblem(Problem):
    """
    The numbers from 1 to 26 are assigned to the letters of the alphabet.
    The numbers beside each word are the total of the values assigned to the letters in the word
    (eg for LYRE: L,Y,R,E might be to equal 5,9,20 and 13 or any other combination that add up to 47).

    Find the value of each letter under the equations:
    BALLET  45     GLEE  66     POLKA      59     SONG     61
    CELLO   43     JAZZ  58     QUARTET    50     SOPRANO  82
    CONCERT 74     LYRE  47     SAXOPHONE 134     THEME    72
    FLUTE   30     OBOE  53     SCALE      51     VIOLIN  100
    FUGUE   50     OPERA 65     SOLO       37     WALTZ    34

    This problem comes from the newsgroup rec.puzzle.
    """

    def __init__(self) -> None:
        """
        Inits the problem.
        """
        super().__init__([(1, 26)] * 26)
        self.add_propagators(
            [
                ([A, B, E, T, L], ALG_AFFINE_EQ, [1, 1, 1, 1, 2, 45]),
                ([C, E, O, L], ALG_AFFINE_EQ, [1, 1, 1, 2, 43]),
                ([E, O, N, R, T, C], ALG_AFFINE_EQ, [1, 1, 1, 1, 1, 2, 74]),
                ([E, F, L, U, T], ALG_AFFINE_EQ, [1, 1, 1, 1, 1, 30]),
                ([E, F, G, U], ALG_AFFINE_EQ, [1, 1, 1, 2, 50]),
                ([G, L, E], ALG_AFFINE_EQ, [1, 1, 2, 66]),
                ([A, J, Z], ALG_AFFINE_EQ, [1, 1, 2, 58]),
                ([E, L, R, Y], ALG_AFFINE_EQ, [1, 1, 1, 1, 47]),
                ([E, B, O], ALG_AFFINE_EQ, [1, 1, 2, 53]),
                ([A, E, P, O, R], ALG_AFFINE_EQ, [1, 1, 1, 1, 1, 65]),
                ([A, K, L, O, P], ALG_AFFINE_EQ, [1, 1, 1, 1, 1, 59]),
                ([A, E, Q, R, U, T], ALG_AFFINE_EQ, [1, 1, 1, 1, 1, 2, 50]),
                ([A, E, H, N, P, S, X, O], ALG_AFFINE_EQ, [1, 1, 1, 1, 1, 1, 1, 2, 134]),
                ([A, C, E, L, S], ALG_AFFINE_EQ, [1, 1, 1, 1, 1, 51]),
                ([L, S, O], ALG_AFFINE_EQ, [1, 1, 2, 37]),
                ([G, N, O, S], ALG_AFFINE_EQ, [1, 1, 1, 1, 61]),
                ([A, N, P, R, S, O], ALG_AFFINE_EQ, [1, 1, 1, 1, 1, 2, 82]),
                ([H, M, T, E], ALG_AFFINE_EQ, [1, 1, 1, 2, 72]),
                ([L, N, O, V, I], ALG_AFFINE_EQ, [1, 1, 1, 1, 2, 100]),
                ([A, L, T, W, Z], ALG_AFFINE_EQ, [1, 1, 1, 1, 1, 34]),
                ([A, B, C, D, E, F, G, H, I, J, K, L, M, N, O, P, Q, R, S, T, U, V, W, X, Y, Z], ALG_ALLDIFFERENT, []),
            ]
        )

    @staticmethod
    def solution_as_dict(solution: NDArray) -> Dict[str, int]:
        """
        Returns the solution as a dict.
        :param solution: the solution as a list of ints
        :return: a dict
        """
        return {
            "A": solution[A],
            "B": solution[B],
            "C": solution[C],
            "D": solution[D],
            "E": solution[E],
            "F": solution[F],
            "G": solution[G],
            "H": solution[H],
            "I": solution[I],
            "J": solution[J],
            "K": solution[K],
            "L": solution[L],
            "M": solution[M],
            "N": solution[N],
            "O": solution[O],
            "P": solution[P],
            "Q": solution[Q],
            "R": solution[R],
            "S": solution[S],
            "T": solution[T],
            "U": solution[U],
            "V": solution[V],
            "W": solution[W],
            "X": solution[X],
            "Y": solution[Y],
            "Z": solution[Z],
        }



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024 - Yan Georget
###############################################################################
import pytest

from nucs.constants import STATS_IDX_SOLVER_SOLUTION_NB
from nucs.examples.schur_lemma.schur_lemma_problem import SchurLemmaProblem
from nucs.solvers.backtrack_solver import BacktrackSolver


class TestSchurLemma:
    @pytest.mark.parametrize(
        "ball_nb, solution_nb",
        [
            (3, 8),
            (4, 17),
            (5, 31),
            (6, 61),
            (7, 124),
            (8, 145),
            (9, 268),
            (10, 147),
            (11, 91),
            (12, 59),
            (13, 8),
            (14, 0),
            (15, 0),
            (16, 0),
            (17, 0),
            (18, 0),
            (19, 0),
            (20, 0),
        ],
    )
    def test_solve(self, ball_nb: int, solution_nb: bool) -> None:
        problem = SchurLemmaProblem(ball_nb)
        solver = BacktrackSolver(problem)
        solver.solve_all()
        assert solver.statistics[STATS_IDX_SOLVER_SOLUTION_NB] == solution_nb



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024 - Yan Georget
###############################################################################
import numpy as np

from nucs.constants import PROP_CONSISTENCY, PROP_ENTAILMENT
from nucs.numpy_helper import new_parameters_by_values, new_shr_domains_by_values
from nucs.propagators.affine_leq_propagator import compute_domains_affine_leq, get_triggers_affine_leq


class TestAffineLEQ:
    def test_get_triggers(self) -> None:
        data = new_parameters_by_values([1, -1, 8])
        assert np.all(get_triggers_affine_leq(2, data) == np.array([1, 2]))

    def test_compute_domains_1(self) -> None:
        domains = new_shr_domains_by_values([(1, 10), (1, 10)])
        data = new_parameters_by_values([1, -1, -1])
        assert compute_domains_affine_leq(domains, data) == PROP_CONSISTENCY
        assert np.all(domains == np.array([[1, 9], [2, 10]]))

    def test_compute_domains_2(self) -> None:
        domains = new_shr_domains_by_values([(1, 10), (1, 10)])
        data = new_parameters_by_values([1, 1, 8])
        assert compute_domains_affine_leq(domains, data) == PROP_CONSISTENCY
        assert np.all(domains == np.array([[1, 7], [1, 7]]))

    def test_compute_domains_3(self) -> None:
        domains = new_shr_domains_by_values([(2, 3), (1, 2)])
        data = new_parameters_by_values([1, 1, 5])
        assert compute_domains_affine_leq(domains, data) == PROP_ENTAILMENT
        assert np.all(domains == np.array([[2, 3], [1, 2]]))



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024 - Yan Georget
###############################################################################
import copy
import logging
from typing import Optional, Self, Tuple, Union

import numpy as np
from numba.typed import List

from nucs.constants import RG_END, RG_START
from nucs.propagators.propagators import GET_COMPLEXITY_FCTS, GET_TRIGGERS_FCTS

logger = logging.getLogger(__name__)


class Problem:
    """
    A problem is defined by a list of domains, a list of variables and a list of propagators.
    A domain is computed by adding an offset to a shared domain.
    """

    def __init__(
        self,
        shr_domains_lst: List[Union[int, Tuple[int, int]]],
        dom_indices_lst: Optional[List[int]] = None,
        dom_offsets_lst: Optional[List[int]] = None,
    ):
        """
        Inits the problem.
        :param shr_domains_lst: the shared domains expressed as a list
        :param dom_indices_lst: the domain indices expressed as a list
        :param dom_offsets_lst: the domain offsets expressed as a list
        """
        n = len(shr_domains_lst)
        if dom_indices_lst is None:
            dom_indices_lst = list(range(0, n))
        if dom_offsets_lst is None:
            dom_offsets_lst = [0] * n
        self.shr_domains_lst = [
            [domain, domain] if isinstance(domain, int) else [domain[0], domain[1]] for domain in shr_domains_lst
        ]
        self.dom_indices_lst = dom_indices_lst
        self.dom_offsets_lst = dom_offsets_lst
        self.shr_domain_nb = len(shr_domains_lst)
        self.propagators: List[Tuple[List[int], int, List[int]]] = []
        self.propagator_nb = 0

    def split(self, split_nb: int, var_idx: int) -> List[Self]:
        """
        Splits a problem into several sub-problems by splitting the domain of a variable.
        :param split_nb: the number of sub-problems
        :param var_idx: the index of the variable
        :return: a list of sub-problems
        """
        logger.debug(f"Splitting in {split_nb} problems with variable {var_idx}")
        shr_dom = self.shr_domains_lst[var_idx]
        shr_dom_min = shr_dom[0]
        shr_dom_max = shr_dom[1]
        shr_dom_sz = shr_dom_max - shr_dom_min + 1
        problems = []
        min_idx = shr_dom_min
        for split_idx in range(split_nb):
            problem = copy.deepcopy(self)
            max_idx = min_idx + shr_dom_sz // split_nb - (0 if split_idx < shr_dom_sz % split_nb else 1)
            problem.shr_domains_lst[var_idx] = [min_idx, max_idx]
            min_idx = max_idx + 1
            problems.append(problem)
        return problems

    def add_variable(
        self, shr_domain: Union[int, Tuple[int, int]], dom_index: Optional[int] = None, dom_offset: Optional[int] = None
    ) -> int:
        """
        Adds an extra variable to the problem.
        :param shr_domain: the shared domain of the variable
        :param dom_index: the domain index (automatically computed if not defined)
        :param dom_offset: the domain offset (set to 0 if not defined)
        :return: the index of the extra variable
        """
        insertion_idx = len(self.shr_domains_lst)
        if dom_index is None:
            dom_index = insertion_idx
        if dom_offset is None:
            dom_offset = 0
        self.shr_domains_lst.append(
            [shr_domain, shr_domain] if isinstance(shr_domain, int) else [shr_domain[0], shr_domain[1]]
        )
        self.dom_indices_lst.append(dom_index)
        self.dom_offsets_lst.append(dom_offset)
        self.shr_domain_nb = len(self.dom_indices_lst)
        return insertion_idx

    def add_variables(
        self,
        shr_domains_list: List[Union[int, Tuple[int, int]]],
        dom_indices_list: Optional[List[int]] = None,
        dom_offsets_list: Optional[List[int]] = None,
    ) -> int:
        """
        Adds extra variables to the problem.
        :param shr_domains_list: the shared domains of the variables
        :param dom_indices_list: the domain indices (automatically computed if not defined)
        :param dom_offsets_list: the domain offsets (set to 0 if not defined)
        :return: the index where the extra variables have been added
        """
        insertion_idx = len(self.shr_domains_lst)
        n = len(shr_domains_list)
        if dom_indices_list is None:
            dom_indices_list = [insertion_idx + i for i in range(n)]
        if dom_offsets_list is None:
            dom_offsets_list = [0] * n
        self.shr_domains_lst.extend(
            [
                [shr_domain, shr_domain] if isinstance(shr_domain, int) else [shr_domain[0], shr_domain[1]]
                for shr_domain in shr_domains_list
            ]
        )
        self.dom_indices_lst.extend(dom_indices_list)
        self.dom_offsets_lst.extend(dom_offsets_list)
        self.shr_domain_nb = len(self.dom_indices_lst)
        return insertion_idx

    def add_propagator(self, propagator: Tuple[List[int], int, List[int]]) -> None:
        """
        Adds an extra propagator.
        :param propagator: the propagator
        """
        self.propagators.append(propagator)
        self.propagator_nb = len(self.propagators)

    def add_propagators(self, propagators: List[Tuple[List[int], int, List[int]]]) -> None:
        """
        Adds a list of propagators
        :param propagators: the propagators
        """
        self.propagators.extend(propagators)
        self.propagator_nb = len(self.propagators)

    def init(self) -> None:
        """
        Completes the initialization of the problem.
        """
        logger.debug("Initializing Problem")
        # Sort the propagators based on their estimated amortized complexities.
        self.propagators.sort(key=lambda prop: GET_COMPLEXITY_FCTS[prop[1]](len(prop[0]), prop[2]))
        # Variable and domain initialization
        self.dom_indices_arr = np.array(self.dom_indices_lst, dtype=np.uint16)
        self.dom_offsets_arr = np.array(self.dom_offsets_lst, dtype=np.int32)
        # Propagator initialization
        self.algorithms = np.empty(self.propagator_nb, dtype=np.uint8)
        # We will store propagator specific data in a global arrays, we need to compute variables and data bounds.
        bound_nb = max(1, self.propagator_nb)
        self.var_bounds = np.zeros((bound_nb, 2), dtype=np.uint16)  # some redundancy here
        self.param_bounds = np.zeros((bound_nb, 2), dtype=np.uint16)  # some redundancy here
        self.var_bounds[0, RG_START] = self.param_bounds[0, RG_START] = 0
        for propagator_idx, propagator in enumerate(self.propagators):
            prop_vars, prop_algorithm, prop_params = propagator
            self.algorithms[propagator_idx] = prop_algorithm
            if propagator_idx > 0:
                self.var_bounds[propagator_idx, RG_START] = self.var_bounds[propagator_idx - 1, RG_END]
                self.param_bounds[propagator_idx, RG_START] = self.param_bounds[propagator_idx - 1, RG_END]
            self.var_bounds[propagator_idx, RG_END] = self.var_bounds[propagator_idx, RG_START] + len(prop_vars)
            self.param_bounds[propagator_idx, RG_END] = self.param_bounds[propagator_idx, RG_START] + len(prop_params)
        # Bounds have been computed and can now be used. The global arrays are the following:
        self.props_dom_indices = np.empty(self.var_bounds[-1, RG_END], dtype=np.uint16)
        self.props_dom_offsets = np.empty(self.var_bounds[-1, RG_END], dtype=np.int32)
        self.props_parameters = np.empty(self.param_bounds[-1, RG_END], dtype=np.int32)
        for propagator_idx, propagator in enumerate(self.propagators):
            prop_vars, _, prop_params = propagator
            var_start = self.var_bounds[propagator_idx, RG_START]
            var_end = self.var_bounds[propagator_idx, RG_END]
            self.props_dom_indices[var_start:var_end] = self.dom_indices_arr[prop_vars]  # cached for faster access
            self.props_dom_offsets[var_start:var_end] = self.dom_offsets_arr[prop_vars]  # cached for faster access
            param_start = self.param_bounds[propagator_idx, RG_START]
            param_end = self.param_bounds[propagator_idx, RG_END]
            self.props_parameters[param_start:param_end] = prop_params
        self.props_dom_offsets = self.props_dom_offsets.reshape((-1, 1))
        self.triggers = np.zeros((self.shr_domain_nb, self.propagator_nb), dtype=np.uint8)
        for propagator_idx, propagator in enumerate(self.propagators):
            prop_vars, prop_algorithm, prop_params = propagator
            triggers = GET_TRIGGERS_FCTS[prop_algorithm](len(prop_vars), prop_params)
            for prop_var_idx, prop_var in enumerate(prop_vars):
                self.triggers[self.dom_indices_arr[prop_var], propagator_idx] = triggers[prop_var_idx]
        logger.debug("Problem initialized")
        logger.info(f"Problem has {self.propagator_nb} propagators")
        logger.info(f"Problem has {self.shr_domain_nb} variables")



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024 - Yan Georget
###############################################################################
import pytest

from nucs.constants import OPTIM_PRUNE, OPTIM_RESET, STATS_LBL_SOLVER_SOLUTION_NB
from nucs.problems.problem import Problem
from nucs.propagators.propagators import ALG_ALLDIFFERENT, ALG_RELATION
from nucs.solvers.backtrack_solver import BacktrackSolver
from nucs.solvers.multiprocessing_solver import MultiprocessingSolver


class TestMultiprocessingSolver:
    def test_solve1(self) -> None:
        problem = Problem([(0, 99), (0, 99)])
        solver = MultiprocessingSolver([BacktrackSolver(problem) for problem in (problem.split(4, 0))])
        solutions = solver.find_all()
        assert len(solutions) == 10000
        statistics = solver.get_statistics()
        assert statistics[STATS_LBL_SOLVER_SOLUTION_NB] == 10000

    def test_solve_alldifferent(self) -> None:
        problem = Problem([(0, 2), (0, 2), (0, 2)])
        problem.add_propagator(([0, 1, 2], ALG_ALLDIFFERENT, []))
        solver = MultiprocessingSolver([BacktrackSolver(problem) for problem in (problem.split(3, 0))])
        solutions = solver.find_all()
        assert len(solutions) == 6
        statistics = solver.get_statistics()
        assert statistics[STATS_LBL_SOLVER_SOLUTION_NB] == 6

    @pytest.mark.parametrize(
        "mode, split_var, split_nb, solution_nb",
        [
            (OPTIM_PRUNE, 0, 1, 6),
            (OPTIM_RESET, 0, 1, 6),
            (OPTIM_PRUNE, 0, 2, 6),
            (OPTIM_RESET, 0, 2, 6),
            (OPTIM_PRUNE, 0, 3, 6),
            (OPTIM_RESET, 0, 3, 6),
            (OPTIM_PRUNE, 1, 1, 6),
            (OPTIM_RESET, 1, 1, 6),
            (OPTIM_PRUNE, 1, 2, 6),
            (OPTIM_RESET, 1, 2, 6),
            (OPTIM_PRUNE, 1, 3, 6),
            (OPTIM_RESET, 1, 3, 6),
        ],
    )
    def test_minimize_relation(self, mode: str, split_var: int, split_nb: int, solution_nb: int) -> None:
        problem = Problem([(-5, 0), (-60, 60)])
        problem.add_propagator(([0, 1], ALG_RELATION, [-5, 25, -4, 16, -3, 9, -2, 4, -1, 1, 0, 0]))
        solver = MultiprocessingSolver([BacktrackSolver(prob) for prob in problem.split(split_nb, split_var)])
        solution = solver.minimize(1, mode=mode)
        assert solution is not None
        assert solution.tolist() == [0, 0]
        statistics = solver.get_statistics()
        assert statistics[STATS_LBL_SOLVER_SOLUTION_NB] == solution_nb



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024 - Yan Georget
###############################################################################
import numpy as np
from numba import njit  # type: ignore
from numpy.typing import NDArray

from nucs.constants import EVENT_MASK_MIN_MAX, MAX, MIN, PROP_CONSISTENCY, PROP_ENTAILMENT, PROP_INCONSISTENCY


def get_complexity_permutation_aux(n: int, parameters: NDArray) -> float:
    """
    Returns the time complexity of the propagator as a float.
    :param n: the number of variables
    :param parameters: the parameters, unused here
    :return: a float
    """
    return n


def get_triggers_permutation_aux(n: int, parameters: NDArray) -> NDArray:
    """
    :param n: the number of variables
    :param parameters: the parameters, unused here
    :return: an array of triggers
    """
    triggers = np.zeros(n, dtype=np.uint8)
    # triggers[:-1] = EVENT_MASK_MIN_MAX  # not worth the cost
    triggers[-1] = EVENT_MASK_MIN_MAX
    return triggers


@njit(cache=True)
def compute_domains_permutation_aux(domains: NDArray, parameters: NDArray) -> int:
    """
    :param domains: the domains of the variables
    :param parameters: the parameters of the propagator
    :return: the status of the propagation (consistency, inconsistency or entailment) as an int
    """
    next = domains[:-1]
    n = len(next)
    prev_j = domains[-1]
    j = parameters[0]
    for i in range(0, n):
        if i < prev_j[MIN] or i > prev_j[MAX]:  # if prev_j != i
            # then next_i != j
            if next[i, MIN] == j:
                next[i, MIN] += 1
            if next[i, MAX] == j:
                next[i, MAX] -= 1
            if next[i, MIN] > next[i, MAX]:
                return PROP_INCONSISTENCY
    if prev_j[MIN] == prev_j[MAX]:
        next[prev_j[MIN]] = j
        return PROP_ENTAILMENT
    return PROP_CONSISTENCY



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024 - Yan Georget
###############################################################################
from nucs.problems.latin_square_problem import M_COLOR, M_COLUMN, M_ROW, LatinSquareRCProblem
from nucs.propagators.propagators import ALG_ELEMENT_LIV


class QuasigroupProblem(LatinSquareRCProblem):
    """
    CSPLIB problem #3 - https://www.csplib.org/Problems/prob003/
    """

    def __init__(self, n: int, symmetry_breaking: bool = True):
        """
        Inits the problem.
        :param n: the size of the quasigroup
        :param symmetry_breaking: a boolean indicating if symmetry constraints should be added to the model
        """
        super().__init__(n)
        # idempotence
        for model in [M_COLOR, M_ROW, M_COLUMN]:
            for i in range(n):
                self.shr_domains_lst[self.cell(i, i, model)] = [i, i]
        if symmetry_breaking:
            for i in range(1, n):
                self.shr_domains_lst[self.cell(i, n - 1)] = [i - 1, n - 1]


class Quasigroup5Problem(QuasigroupProblem):
    """
    Defined by:
       ((b∗a)∗b)∗b=a
    Can be enforced by:
       color[color[j, i] ,j] = row[i, j]
    which avoids the creation of additional variables
    """

    def __init__(self, n: int, symmetry_breaking: bool = True):
        """
        Inits the problem
        :param n: the size of the quasigroup
        :param symmetry_breaking: a boolean indicating if symmetry constraints should be added to the model
        """
        super().__init__(n, symmetry_breaking)
        for j in range(n):
            for i in range(n):
                if i != j:
                    self.add_propagator(
                        (
                            [*self.column(j, M_COLOR), self.cell(j, i, M_COLOR), self.cell(i, j, M_ROW)],
                            ALG_ELEMENT_LIV,
                            [],
                        )
                    )



from numba import njit  # type: ignore
from numpy.typing import NDArray

from nucs.constants import DOM_UPDATE_EVENTS, DOM_UPDATE_IDX, MAX, MIN, STATS_IDX_SOLVER_BACKTRACK_NB
from nucs.propagators.propagators import update_propagators


@njit(cache=True)
def cp_init(
    shr_domains_stack: NDArray,
    not_entailed_propagators_stack: NDArray,
    dom_update_stack: NDArray,
    stacks_top: NDArray,
    shr_domains_arr: NDArray,
) -> None:
    """
    Inits the choice points.
    :param shr_domains_stack: the stack of shared domains
    :param not_entailed_propagators_stack: the stack of not entailed propagators
    :param dom_update_stack: the stack of domain updates
    :param stacks_top: the index of the top of the stacks as a Numpy array
    :param shr_domains_arr: the shared domains
    :return: a Numpy array
    """
    shr_domains_stack[0] = shr_domains_arr
    not_entailed_propagators_stack[0] = True
    dom_update_stack[:, :] = 0
    stacks_top[0] = 0


@njit(cache=True)
def cp_put(shr_domains_stack: NDArray, not_entailed_propagators_stack: NDArray, stacks_top: NDArray) -> None:
    """
    Adds a choice point to the stack of choice points.
    :param shr_domains_stack: the stack of shared domains
    :param not_entailed_propagators_stack: the stack of not entailed propagators
    :param stacks_top: the index of the top of the stacks as a Numpy array
    """
    cp_top_idx = stacks_top[0]
    shr_domains_stack[cp_top_idx + 1, :, :] = shr_domains_stack[cp_top_idx, :, :]
    not_entailed_propagators_stack[cp_top_idx + 1, :] = not_entailed_propagators_stack[cp_top_idx, :]
    stacks_top[0] = cp_top_idx + 1


@njit(cache=True)
def backtrack(
    statistics: NDArray,
    not_entailed_propagators_stack: NDArray,
    dom_update_stack: NDArray,
    stacks_top: NDArray,
    triggered_propagators: NDArray,
    triggers: NDArray,
) -> bool:
    """
    Backtracks and updates the problem's domains.
    :param statistics: the statistics array
    :param not_entailed_propagators_stack: the stack of not entailed propagators
    :param dom_update_stack: the stack of domain updates
    :param stacks_top: the index of the top of the stacks as a Numpy array
    :param triggered_propagators: the set propagators that are currently triggered as a Numpy array
    :param triggers: a Numpy array of event masks indexed by shared domain indices and propagators
    :return: true iff it is possible to backtrack
    """
    if stacks_top[0] == 0:
        return False
    stacks_top[0] -= 1
    statistics[STATS_IDX_SOLVER_BACKTRACK_NB] += 1
    update_propagators(
        triggered_propagators,
        not_entailed_propagators_stack[stacks_top[0]],
        triggers,
        dom_update_stack[stacks_top[0], DOM_UPDATE_IDX],
        dom_update_stack[stacks_top[0], DOM_UPDATE_EVENTS],
    )
    return True


@njit(cache=True)
def fix_choice_points(
    shr_domains_stack: NDArray,
    stacks_top: NDArray,
    dom_indices_arr: NDArray,
    dom_offsets_arr: NDArray,
    variable_idx: int,
    value: int,
    bound: int,
) -> bool:
    """
    :param problem: the problem
    :param shr_domains_stack: the stack of shared domains
    :param not_entailed_propagators_stack: the stack of not entailed propagators
    :param dom_update_stack: the stack of domain updates
    :param stacks_top: the index of the top of the stacks as a Numpy array
    :param triggered_propagators: the list of triggered propagators
    """
    dom_idx = dom_indices_arr[variable_idx]
    shr_domains_stack[0 : stacks_top[0] + 1, dom_idx, bound] = (
        value + (1 if bound == MIN else -1) - dom_offsets_arr[variable_idx]
    )
    while shr_domains_stack[stacks_top[0], dom_idx, MIN] > shr_domains_stack[stacks_top[0], dom_idx, MAX]:
        if stacks_top[0] == 0:
            return False
        stacks_top[0] -= 1
    return True


@njit(cache=True)
def fix_top_choice_point(
    shr_domains_stack: NDArray,
    stacks_top: NDArray,
    dom_indices_arr: NDArray,
    dom_offsets_arr: NDArray,
    variable_idx: int,
    value: int,
    bound: int,
) -> bool:
    """
    :param problem: the problem
    :param shr_domains_stack: the stack of shared domains
    :param not_entailed_propagators_stack: the stack of not entailed propagators
    :param dom_update_stack: the stack of domain updates
    :param stacks_top: the index of the top of the stacks as a Numpy array
    :param triggered_propagators: the list of triggered propagators
    """
    dom_idx = dom_indices_arr[variable_idx]
    shr_domains_stack[stacks_top[0], dom_idx, bound] = (
        value + (1 if bound == MIN else -1) - dom_offsets_arr[variable_idx]
    )
    while shr_domains_stack[stacks_top[0], dom_idx, MIN] > shr_domains_stack[stacks_top[0], dom_idx, MAX]:
        if stacks_top[0] == 0:
            return False
        stacks_top[0] -= 1
    return True



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024 - Yan Georget
###############################################################################
import sys

import numpy as np
from numba import njit  # type: ignore
from numpy.typing import NDArray

from nucs.constants import EVENT_MASK_MIN_MAX, MAX, MIN, PROP_CONSISTENCY, PROP_INCONSISTENCY


def get_complexity_total_cost(n: int, parameters: NDArray) -> float:
    """
    Returns the time complexity of the propagator as a float.
    :param n: the number of variables
    :param parameters: the parameters, unused here
    :return: a float
    """
    return n * n


def get_triggers_total_cost(n: int, parameters: NDArray) -> NDArray:
    """
    This propagator is triggered whenever there is a change in the domain of a variable.
    :param n: the number of variables
    :param parameters: the parameters, unused here
    :return: an array of triggers
    """
    triggers = np.full(n, dtype=np.uint8, fill_value=EVENT_MASK_MIN_MAX)
    triggers[-1] = 0
    return triggers


@njit(cache=True)
def compute_domains_total_cost(domains: NDArray, parameters: NDArray) -> int:
    """
    :param domains: the domains of the variables
    :param parameters: the parameters of the propagator
    :return: the status of the propagation (consistency, inconsistency or entailment) as an int
    """
    n = len(domains) - 1
    used = np.zeros(n, dtype=np.bool)
    for i in range(n):
        if domains[i, MIN] == domains[i, MAX]:
            used[domains[i, MIN]] = True
    global_min = 0
    # for j in range(n):
    #     max_regret = 0
    #     min_cost = min([costs[i, j] for i in range(n) if costs[i,j] > 0])
    #     for i in range(n):
    #         if costs[i, j] == min_cost:
    #             regret = compute_regret(domains[i], costs[i])
    #             if regret > max_regret:
    #                 max_regret = regret
    #             global_min += regret
    #     global_min -= max_regret
    global_max = 0
    for i in range(n):
        if domains[i, MIN] == domains[i, MAX]:
            local_min = local_max = parameters[i * n + domains[i, MIN]]
        else:
            local_min = sys.maxsize
            local_max = 0
            for value in range(domains[i, MIN], domains[i, MAX] + 1):
                if not used[value]:
                    cost = parameters[i * n + value]
                    if cost > 0:
                        if cost < local_min:
                            local_min = cost
                        if cost > local_max:
                            local_max = cost
            # if local_min > local_max:
            #     return PROP_INCONSISTENCY
        global_min += local_min
        global_max += local_max
    domains[-1, MIN] = max(domains[-1, MIN], global_min)
    domains[-1, MAX] = min(domains[-1, MAX], global_max)
    if domains[-1, MIN] > domains[-1, MAX]:
        return PROP_INCONSISTENCY
    return PROP_CONSISTENCY



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024 - Yan Georget
###############################################################################
import sys

from numba import njit  # type: ignore
from numpy.typing import NDArray

from nucs.constants import MAX, MIN
from nucs.heuristics.max_regret_var_heuristic import regret


@njit(cache=True)
def tsp_var_heuristic(
    decision_domains: NDArray, shr_domains_stack: NDArray, stacks_top: NDArray, params: NDArray
) -> int:
    """
    :param decision_domains: the indices of a subset of the shared domains
    :param shr_domains_stack: the stack of shared domains
    :param stacks_top: the index of the top of the stacks as a Numpy array
    :param params: a two-dimensional (first dimension correspond to variables, second to values) costs array
    :return: the index of the shared domain
    """
    best_score = -sys.maxsize
    best_idx = -1
    top = stacks_top[0]
    for dom_idx in decision_domains:
        shr_domain = shr_domains_stack[top, dom_idx]
        if 0 < shr_domain[MAX] - shr_domain[MIN]:
            score = compute_score(shr_domain, dom_idx, params)
            if best_score < score:
                best_idx = dom_idx
                best_score = score
    return best_idx


@njit(cache=True)
def compute_score(shr_domain: NDArray, dom_idx: int, params: NDArray) -> int:
    """
    Minimize [min(5, size(X)), -regret(X)] for lexicographic order.
    """
    size = min(12, shr_domain[MAX] - shr_domain[MIN] + 1)
    return -size * 1024 + regret(shr_domain, params[dom_idx])



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024 - Yan Georget
###############################################################################
import argparse

from rich import print

from nucs.constants import LOG_LEVEL_INFO, LOG_LEVELS
from nucs.examples.queens.queens_problem import QueensProblem
from nucs.heuristics.heuristics import VAR_HEURISTIC_FIRST_NOT_INSTANTIATED, VAR_HEURISTIC_SMALLEST_DOMAIN
from nucs.solvers.backtrack_solver import BacktrackSolver
from nucs.solvers.consistency_algorithms import CONSISTENCY_ALG_BC, CONSISTENCY_ALG_SHAVING
from nucs.solvers.multiprocessing_solver import MultiprocessingSolver

# Run with the following command (the second run is much faster because the code has been compiled):
# NUMBA_CACHE_DIR=.numba/cache python -m nucs.examples.queens -n 10
if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--ff", type=bool, action=argparse.BooleanOptionalAction, default=False)
    parser.add_argument("--log_level", choices=LOG_LEVELS, default=LOG_LEVEL_INFO)
    parser.add_argument("-n", type=int, default=10)
    parser.add_argument("--processors", type=int, default=1)
    parser.add_argument("--shaving", type=bool, action=argparse.BooleanOptionalAction, default=False)
    args = parser.parse_args()
    problem = QueensProblem(args.n)
    solver = (
        MultiprocessingSolver(
            [
                BacktrackSolver(
                    problem,
                    consistency_alg_idx=CONSISTENCY_ALG_SHAVING if args.shaving else CONSISTENCY_ALG_BC,
                    var_heuristic_idx=(
                        VAR_HEURISTIC_SMALLEST_DOMAIN if args.ff else VAR_HEURISTIC_FIRST_NOT_INSTANTIATED
                    ),
                    log_level=args.log_level,
                )
                for problem in problem.split(args.processors, 0)
            ]
        )
        if args.processors > 1
        else BacktrackSolver(
            problem,
            consistency_alg_idx=CONSISTENCY_ALG_SHAVING if args.shaving else CONSISTENCY_ALG_BC,
            var_heuristic_idx=VAR_HEURISTIC_SMALLEST_DOMAIN if args.ff else VAR_HEURISTIC_FIRST_NOT_INSTANTIATED,
            log_level=args.log_level,
        )
    )
    solver.solve_all()
    print(solver.get_statistics())



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024 - Yan Georget
###############################################################################
from nucs.problems.problem import Problem
from nucs.propagators.propagators import ALG_ALLDIFFERENT, ALG_PERMUTATION_AUX


class PermutationProblem(Problem):
    """
    A model for permutation.
    """

    def __init__(self, n: int):
        """
        Inits the permutation problem.
        :param n: the number variables/values
        """
        self.n = n
        shr_domains = [(0, n - 1)] * 2 * n
        super().__init__(shr_domains)
        for i in range(n):
            self.add_propagator((list(range(n)) + [n + i], ALG_PERMUTATION_AUX, [i]))
            self.add_propagator((list(range(n, 2 * n)) + [i], ALG_PERMUTATION_AUX, [i]))
        self.add_propagator((list(range(n)), ALG_ALLDIFFERENT, []))
        self.add_propagator((list(range(n, 2 * n)), ALG_ALLDIFFERENT, []))



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024 - Yan Georget
###############################################################################
import math

import numpy as np
from numba import njit  # type: ignore
from numpy.typing import NDArray

from nucs.constants import EVENT_MASK_GROUND, EVENT_MASK_MIN, MAX, MIN
from nucs.heuristics.heuristics import first_not_instantiated_var_heuristic
from nucs.problems.problem import Problem
from nucs.propagators.propagators import ALG_AFFINE_EQ, ALG_AFFINE_LEQ, ALG_ALLDIFFERENT, update_propagators
from nucs.solvers.bound_consistency_algorithm import bound_consistency_algorithm

GOLOMB_LENGTHS = np.array([0, 0, 1, 3, 6, 11, 17, 25, 34, 44, 55, 72, 85, 106, 127])


@njit(cache=True)
def sum_first(n: int) -> int:
    """
    Returns the sum of the first n integers.
    :param n: an integer
    :return: the sum
    """
    return (n * (n + 1)) // 2


@njit(cache=True)
def index(mark_nb: int, i: int, j: int) -> int:
    """
    Returns the index of the distance variable between two marks.
    :param mark_nb: the total number of marks
    :param i: the first mark
    :param j: the second mark
    :return: the index of the distance variable
    """
    return i * mark_nb - sum_first(i) + j - i - 1


class GolombProblem(Problem):
    """
    This is the famous Golomb ruler problem.

    It consists in finding n integers mark_i such that:
    - mark_0 = 0,
    - mark_0 <...< mark_n-1,
    - for all i < j, mark_j - mark_i are different,
    - mark_n-1 is minimal.

    CSPLIB problem #6 - https://www.csplib.org/Problems/prob006/
    """

    def __init__(self, mark_nb: int, symmetry_breaking: bool = True) -> None:
        dist_nb = sum_first(mark_nb - 1)
        domains = np.empty((dist_nb, 2), dtype=np.int32)
        for i in range(0, mark_nb - 1):
            for j in range(i + 1, mark_nb):
                domains[index(mark_nb, i, j), MIN] = (
                    GOLOMB_LENGTHS[j - i + 1] if j - i + 1 < mark_nb else sum_first(j - i)
                )
        domains[:, MAX] = sum_first(dist_nb) - sum_first(dist_nb - mark_nb)
        super().__init__(domains)
        self.length_idx = index(mark_nb, 0, mark_nb - 1)  # we want to minimize this
        # dist_ij = mark_j - mark_i for j > i
        # mark_j = dist_0j for j > 0
        for i in range(1, mark_nb - 1):
            for j in range(i + 1, mark_nb):
                self.add_propagator(
                    (
                        [index(mark_nb, 0, j), index(mark_nb, 0, i), index(mark_nb, i, j)],
                        ALG_AFFINE_EQ,
                        [1, -1, -1, 0],
                    )
                )
        self.add_propagator((list(range(self.shr_domain_nb)), ALG_ALLDIFFERENT, []))
        # redundant constraints
        for i in range(mark_nb - 1):
            for j in range(i + 1, mark_nb):
                if j - i < mark_nb - 1:
                    self.add_propagator(
                        (
                            [index(mark_nb, i, j), index(mark_nb, 0, mark_nb - 1)],
                            ALG_AFFINE_LEQ,
                            [1, -1, -sum_first(mark_nb - 1 - (j - i))],
                        )
                    )
        if symmetry_breaking:
            self.add_propagator(
                (
                    [index(mark_nb, 0, 1), index(mark_nb, mark_nb - 2, mark_nb - 1)],
                    ALG_AFFINE_LEQ,
                    [1, -1, -1],
                )
            )


@njit(cache=True)
def golomb_consistency_algorithm(
    statistics: NDArray,
    algorithms: NDArray,
    var_bounds: NDArray,
    param_bounds: NDArray,
    dom_indices_arr: NDArray,
    dom_offsets_arr: NDArray,
    props_dom_indices: NDArray,
    props_dom_offsets: NDArray,
    props_parameters: NDArray,
    shr_domains_propagators: NDArray,
    shr_domains_stack: NDArray,
    not_entailed_propagators_stack: NDArray,
    dom_update_stack: NDArray,
    stacks_top: NDArray,
    triggered_propagators: NDArray,
    compute_domains_addrs: NDArray,
    decision_domains: NDArray,
) -> int:
    """
    Applies a custom consistency algorithm for the Golomb Ruler problem.
    :param statistics: the statistics array
    :param problem: the problem
    :return: the status as an int
    """
    top = stacks_top[0]
    # first prune the search space
    mark_nb = (1 + int(math.sqrt(8 * len(dom_indices_arr) + 1))) // 2
    ni_var_idx = first_not_instantiated_var_heuristic(
        decision_domains, shr_domains_stack, stacks_top, None
    )  # no domains shared between vars
    if 1 < ni_var_idx < mark_nb - 1:  # otherwise useless
        used_distance = np.zeros(sum_first(mark_nb - 2) + 1, dtype=np.bool)
        # a reusable array for storing the minimal sum of different integers:
        # minimal_sum[i] will be the minimal sum of i different integers chosen among a set of possible integers
        minimal_sum = np.zeros(mark_nb - 1, dtype=np.int32)
        # the following will mark at most sum(n-3) numbers as used
        # hence there will be at least n-2 unused numbers greater than 0
        for var_idx in range(index(mark_nb, ni_var_idx - 2, ni_var_idx - 1) + 1):
            dist = shr_domains_stack[top, dom_indices_arr[var_idx], MIN]  # no offset
            if dist < len(used_distance):
                used_distance[dist] = True
        # let's compute the sum of non-used numbers
        distance = 1
        for j in range(0, mark_nb - ni_var_idx):
            while used_distance[distance]:
                distance += 1
            minimal_sum[j + 1] = minimal_sum[j] + distance
            distance += 1
        for i in range(ni_var_idx - 1, mark_nb - 1):
            for j in range(i + 1, mark_nb):
                dom_idx = dom_indices_arr[index(mark_nb, i, j)]
                shr_domains_stack[top, dom_idx, MIN] = minimal_sum[j - i]  # no offset
                events = EVENT_MASK_MIN
                if shr_domains_stack[top, dom_idx, MIN] == shr_domains_stack[top, dom_idx, MAX]:
                    events |= EVENT_MASK_GROUND
                update_propagators(
                    triggered_propagators,
                    not_entailed_propagators_stack[top],
                    shr_domains_propagators,
                    dom_idx,
                    events,
                )
    return bound_consistency_algorithm(
        statistics,
        algorithms,
        var_bounds,
        param_bounds,
        dom_indices_arr,
        dom_offsets_arr,
        props_dom_indices,
        props_dom_offsets,
        props_parameters,
        shr_domains_propagators,
        shr_domains_stack,
        not_entailed_propagators_stack,
        dom_update_stack,
        stacks_top,
        triggered_propagators,
        compute_domains_addrs,
        decision_domains,
    )



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024 - Yan Georget
###############################################################################
import numpy as np

from nucs.constants import PROP_CONSISTENCY
from nucs.numpy_helper import new_parameters_by_values, new_shr_domains_by_values
from nucs.propagators.alldifferent_propagator import compute_domains_alldifferent, path_max, path_min, path_set


class TestAlldifferent:
    def test_path_min(self) -> None:
        a = np.array([2, 3, 4, 0, 1])
        assert path_min(a, 0) == 0
        assert path_min(a, 1) == 1
        assert path_min(a, 2) == 2
        assert path_min(a, 3) == 0
        assert path_min(a, 4) == 1

    def test_path_max(self) -> None:
        a = np.array([2, 3, 4, 0, 1])
        assert path_max(a, 0) == 4
        assert path_max(a, 1) == 3
        assert path_max(a, 2) == 4
        assert path_max(a, 3) == 3
        assert path_max(a, 4) == 4

    def test_path_set(self) -> None:
        a = np.array([2, 3, 4, 0, 1])
        path_set(a, 0, 4, -1)
        assert np.all(a == np.array([-1, 3, -1, 0, 1]))

    def test_compute_domains_1(self) -> None:
        domains = new_shr_domains_by_values([(3, 6), (3, 4), (2, 5), (2, 4), (3, 4), (1, 6)])
        data = new_parameters_by_values([])
        assert compute_domains_alldifferent(domains, data) == PROP_CONSISTENCY
        assert np.all(domains == np.array([[6, 6], [3, 4], [5, 5], [2, 2], [3, 4], [1, 1]]))

    def test_compute_domains_2(self) -> None:
        domains = new_shr_domains_by_values([(0, 0), (2, 2), (1, 2)])
        data = new_parameters_by_values([])
        assert compute_domains_alldifferent(domains, data) == PROP_CONSISTENCY
        assert np.all(domains == np.array([[0, 0], [2, 2], [1, 1]]))

    def test_compute_domains_3(self) -> None:
        domains = new_shr_domains_by_values([(0, 0), (0, 4), (0, 4), (0, 4), (0, 4)])
        data = new_parameters_by_values([])
        assert compute_domains_alldifferent(domains, data) == PROP_CONSISTENCY
        assert np.all(domains == np.array([[0, 0], [1, 4], [1, 4], [1, 4], [1, 4]]))



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024 - Yan Georget
###############################################################################
import numpy as np
from numba import njit  # type: ignore
from numpy.typing import NDArray

from nucs.constants import EVENT_MASK_GROUND, MAX, MIN, PROP_CONSISTENCY, PROP_INCONSISTENCY

PATH_START = 0
PATH_END = 1
PATH_LENGTH = 2


def get_complexity_no_sub_cycle(n: int, parameters: NDArray) -> float:
    """
    Returns the time complexity of the propagator as a float.
    :param n: the number of variables
    :param parameters: the parameters, unused here
    :return: a float
    """
    return n * n


def get_triggers_no_sub_cycle(n: int, parameters: NDArray) -> NDArray:
    """
    Returns the triggers for this propagator.
    :param n: the number of variables
    :param parameters: the parameters, unused here
    :return: an array of triggers
    """
    return np.full(n, dtype=np.uint8, fill_value=EVENT_MASK_GROUND)


@njit(cache=True)
def compute_domains_no_sub_cycle(domains: NDArray, parameters: NDArray) -> int:
    """
    :param domains: the domains of the variables
    :param parameters: unused here
    :return: the status of the propagation (consistency, inconsistency or entailment) as an int
    """
    n = len(domains)
    paths = np.zeros((n, 3), dtype=np.int16)
    for i in range(n):
        paths[i, :PATH_LENGTH] = i
    loop = True
    while loop:
        loop = False
        for i in range(n):
            if domains[i, MIN] == domains[i, MAX]:
                j = domains[i, MIN]
                if i == j:
                    return PROP_INCONSISTENCY
                if paths[i, PATH_END] == i:
                    end = paths[i, PATH_END] = paths[j, PATH_END]
                    start = paths[j, PATH_START] = paths[i, PATH_START]
                    paths[start, PATH_END] = end
                    paths[end, PATH_START] = start
                    length = paths[i, PATH_LENGTH] + 1 + paths[j, PATH_LENGTH]
                    paths[i, PATH_LENGTH] = paths[j, PATH_LENGTH] = paths[start, PATH_LENGTH] = paths[
                        end, PATH_LENGTH
                    ] = length
                    if length < n - 1:
                        if domains[end, MIN] == start:
                            domains[end, MIN] = start + 1
                        if domains[end, MAX] == start:
                            domains[end, MAX] = start - 1
                        if domains[end, MIN] > domains[end, MAX]:
                            return PROP_INCONSISTENCY
                        if end < i:
                            loop = True
    return PROP_CONSISTENCY



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024 - Yan Georget
###############################################################################
from typing import List

from nucs.problems.problem import Problem
from nucs.propagators.propagators import ALG_ALLDIFFERENT


class QueensProblem(Problem):
    """
    A simple model for the n-queens problem.

    CSPLIB problem #54 - https://www.csplib.org/Problems/prob054/
    """

    def __init__(self, n: int):
        """
        Inits the problem.
        :param n: the number of queens
        """
        super().__init__(
            [(0, n - 1)] * n,
            list(range(n)) * 3,
            [0] * n + list(range(n)) + list(range(0, -n, -1)),
        )
        self.add_propagator((list(range(n)), ALG_ALLDIFFERENT, []))
        self.add_propagator((list(range(n, 2 * n)), ALG_ALLDIFFERENT, []))
        self.add_propagator((list(range(2 * n, 3 * n)), ALG_ALLDIFFERENT, []))

    @staticmethod
    def solution_as_matrix(solution: List[int]) -> List[List[str]]:
        n = len(solution)
        return [([" "] * i + ["X"] + [" "] * (n - i - 1)) for i in range(n)]



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024 - Yan Georget
###############################################################################
import pytest

from nucs.constants import STATS_IDX_ALG_BC_NB, STATS_IDX_SOLVER_CHOICE_NB, STATS_IDX_SOLVER_SOLUTION_NB
from nucs.problems.circuit_problem import CircuitProblem
from nucs.solvers.backtrack_solver import BacktrackSolver


class TestCircuitProblem:
    @pytest.mark.parametrize("size, solution_nb", [(2, 1), (3, 2), (4, 6)])
    def test_solve_all(self, size: int, solution_nb: int) -> None:
        problem = CircuitProblem(size)
        solver = BacktrackSolver(problem)
        solver.solve_all()
        assert solver.statistics[STATS_IDX_SOLVER_SOLUTION_NB] == solution_nb

    def test_no_sub_cycle_5(self) -> None:
        problem = CircuitProblem(5)
        problem.shr_domains_lst[1] = [2, 2]
        problem.shr_domains_lst[2] = [1, 1]
        solver = BacktrackSolver(problem)
        solver.solve_all()
        assert solver.statistics[STATS_IDX_SOLVER_SOLUTION_NB] == 0
        assert solver.statistics[STATS_IDX_SOLVER_CHOICE_NB] == 0
        assert solver.statistics[STATS_IDX_ALG_BC_NB] == 1



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024 - Yan Georget
###############################################################################
import pytest

from nucs.constants import STATS_IDX_SOLVER_SOLUTION_NB
from nucs.examples.magic_square.magic_square_problem import MagicSquareProblem
from nucs.heuristics.heuristics import DOM_HEURISTIC_MAX_VALUE, VAR_HEURISTIC_SMALLEST_DOMAIN
from nucs.solvers.backtrack_solver import BacktrackSolver


class TestMagicSquare:

    def test_first_diagonal(self) -> None:
        assert MagicSquareProblem(3).first_diag() == [0, 4, 8]

    def test_second_diagonal(self) -> None:
        assert MagicSquareProblem(3).second_diag() == [6, 4, 2]

    @pytest.mark.parametrize("size,solution_nb", [(2, 0), (3, 1), (4, 880)])
    def test_magic_square(self, size: int, solution_nb: int) -> None:
        problem = MagicSquareProblem(size)
        solver = BacktrackSolver(
            problem, var_heuristic_idx=VAR_HEURISTIC_SMALLEST_DOMAIN, dom_heuristic_idx=DOM_HEURISTIC_MAX_VALUE
        )
        solver.solve_all()
        assert solver.statistics[STATS_IDX_SOLVER_SOLUTION_NB] == solution_nb



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024 - Yan Georget
###############################################################################
import numpy as np

from nucs.constants import PROP_CONSISTENCY
from nucs.numpy_helper import new_parameters_by_values, new_shr_domains_by_values
from nucs.propagators.affine_eq_propagator import compute_domains_affine_eq, get_triggers_affine_eq


class TestAffineEQ:
    def test_get_triggers(self) -> None:
        data = new_parameters_by_values([8, 1, -1])
        assert np.all(get_triggers_affine_eq(2, data) == np.array([3, 3]))

    def test_compute_domains_1(self) -> None:
        domains = new_shr_domains_by_values([(1, 10), (1, 10)])
        data = new_parameters_by_values([1, 1, 8])
        assert compute_domains_affine_eq(domains, data) == PROP_CONSISTENCY
        assert np.all(domains == np.array([[1, 7], [1, 7]]))

    def test_compute_domains_2(self) -> None:
        domains = new_shr_domains_by_values([(5, 10), (5, 10), (5, 10)])
        data = new_parameters_by_values([1, 1, 1, 27])
        assert compute_domains_affine_eq(domains, data) == PROP_CONSISTENCY
        assert np.all(domains == np.array([[7, 10], [7, 10], [7, 10]]))

    def test_compute_domains_3(self) -> None:
        domains = new_shr_domains_by_values([(-2, -1), (2, 3)])
        data = new_parameters_by_values([1, 1, 0])
        assert compute_domains_affine_eq(domains, data) == PROP_CONSISTENCY
        assert np.all(domains == np.array([[-2, -2], [2, 2]]))

    def test_compute_domains_4(self) -> None:
        domains = new_shr_domains_by_values([(1, 10), (1, 10)])
        data = new_parameters_by_values([1, -3, 0])
        assert compute_domains_affine_eq(domains, data) == PROP_CONSISTENCY
        assert np.all(domains == np.array([[3, 10], [1, 3]]))

    def test_compute_domains_5(self) -> None:
        domains = new_shr_domains_by_values([(-14, 11), (-4, 5)])
        data = new_parameters_by_values([1, 3, 0])
        assert compute_domains_affine_eq(domains, data) == PROP_CONSISTENCY
        assert np.all(domains == np.array([[-14, 11], [-3, 4]]))

    def test_compute_domains_6(self) -> None:
        domains = new_shr_domains_by_values([4, 3, 5, 9, 1, 8, 6, 2, 7, 0])
        data = new_parameters_by_values([200, -1000, 100002, 9900, 100000, 20, 1000, 0, -99010, -1, 0])
        assert compute_domains_affine_eq(domains, data) == PROP_CONSISTENCY



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024 - Yan Georget
###############################################################################
from typing import List

from nucs.problems.problem import Problem
from nucs.propagators.propagators import ALG_AND, ALG_EXACTLY_TRUE, ALG_LEXICOGRAPHIC_LEQ


class BIBDProblem(Problem):
    """
    CSPLIB problem #28 - https://www.csplib.org/Problems/prob028/
    """

    def __init__(self, v: int, b: int, r: int, k: int, l: int, symmetry_breaking: bool = True):
        """
        Inits the problem.
        :param v: the number of points/rows
        :param b: the number of blocks
        :param r: the number of true values per row
        :param k: the number of true values per column
        :param l: the scalar product between two rows
        :param symmetry_breaking: a boolean indicating if symmetry constraints should be added to the model
        """
        self.v = v  # number of points/rows
        self.b = b  # number of blocks/columns
        matrix_var_nb = v * b  # number of cells in the matrix
        additional_var_nb = ((v * (v - 1)) // 2) * b
        super().__init__([(0, 1)] * (matrix_var_nb + additional_var_nb))
        # rows: counts
        for object_idx in range(0, v):
            self.add_propagator((list(range(object_idx * b, (object_idx + 1) * b)), ALG_EXACTLY_TRUE, [r]))
        # columns: counts
        for block_idx in range(0, b):
            self.add_propagator((list(range(block_idx, v * b, b)), ALG_EXACTLY_TRUE, [k]))
        # scalar products: conjunctions and counts
        conj_idx = v * b  # index of first redundant variable
        for i1 in range(0, v - 1):
            for i2 in range(i1 + 1, v):
                conj_vars = []
                for block_idx in range(0, b):
                    self.add_propagator(([i1 * b + block_idx, i2 * b + block_idx, conj_idx], ALG_AND, []))
                    conj_vars.append(conj_idx)
                    conj_idx += 1
                self.add_propagator((conj_vars, ALG_EXACTLY_TRUE, [l]))
        if symmetry_breaking:
            # lexleq on rows
            for object_idx in range(0, v - 1):
                self.add_propagator((list(range(object_idx * b, (object_idx + 2) * b)), ALG_LEXICOGRAPHIC_LEQ, []))
            # lexleq on columns
            for block_idx in range(0, b - 1):
                self.add_propagator(
                    (
                        list(range(block_idx, v * b, b)) + list(range(block_idx + 1, v * b, b)),
                        ALG_LEXICOGRAPHIC_LEQ,
                        [],
                    )
                )

    def solution_as_matrix(self, solution: List[int]) -> List[List[int]]:
        """
        Returns the solutions as a matrix of ints.
        :param solution: the solution as a list of ints
        :return: a matrix
        """
        return [[solution[i * self.b + j] for j in range(0, self.b)] for i in range(0, self.v)]



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024 - Yan Georget
###############################################################################
from typing import List

from numba import types  # type: ignore
from numba.core import cgutils
from numba.experimental.function_type import _get_wrapper_address
from numba.extending import intrinsic


@intrinsic
def function_from_address(typingctx, func_type_ref: types.FunctionType, addr: int):  # type: ignore
    """
    Recovers a function from FunctionType and address.
    """
    func_type = func_type_ref.instance_type

    def codegen(context, builder, sig, args):  # type: ignore
        _, address = args
        sfunc = cgutils.create_struct_proxy(func_type)(context, builder)
        sfunc.addr = builder.inttoptr(address, context.get_value_type(types.voidptr))
        return sfunc._getvalue()

    return func_type(func_type_ref, addr), codegen


def build_function_address_list(functions, signature) -> List[int]:  # type: ignore
    return [_get_wrapper_address(function, signature) for function in functions]
